@@ -561,7 +561,7 @@ int initgroups(const char *name, gid_t basegid)
     return setgroups(index, groups);
 #endif /* def QNX */
 }
-#endif /* def NEED_INITGROUPS */
+#endif /* def HAVE_INITGROUPS */
 
 #ifdef AP_MPM_USES_POD
 
@@ -693,13 +693,15 @@ static apr_status_t dummy_connection(ap_pod_t *pod)
 
         ap_log_error(APLOG_MARK, log_level, rv, ap_server_conf,
                      "connect to listener on %pI", ap_listeners->bind_addr);
+        apr_pool_destroy(p);
+        return rv;
     }
 
     /* Create the request string. We include a User-Agent so that
      * adminstrators can track down the cause of the odd-looking
      * requests in their logs.
      */
-    srequest = apr_pstrcat(p, "GET / HTTP/1.0\r\nUser-Agent: ",
+    srequest = apr_pstrcat(p, "OPTIONS * HTTP/1.0\r\nUser-Agent: ",
                            ap_get_server_banner(),
                            " (internal dummy connection)\r\n\r\n", NULL);
 
@@ -965,7 +967,6 @@ int ap_signal_server(int *exit_status, apr_pool_t *pconf)
     apr_status_t rv;
     pid_t otherpid;
     int running = 0;
-    int have_pid_file = 0;
     const char *status;
 
     *exit_status = 0;
@@ -983,7 +984,6 @@ int ap_signal_server(int *exit_status, apr_pool_t *pconf)
         status = "httpd (no pid file) not running";
     }
     else {
-        have_pid_file = 1;
         if (kill(otherpid, 0) == 0) {
             running = 1;
             status = apr_psprintf(pconf,
@@ -1058,12 +1058,10 @@ void ap_mpm_rewrite_args(process_rec *process)
     apr_getopt_t *opt;
     char optbuf[3];
     const char *optarg;
-    int fixed_args;
 
     mpm_new_argv = apr_array_make(process->pool, process->argc,
                                   sizeof(const char **));
     *(const char **)apr_array_push(mpm_new_argv) = process->argv[0];
-    fixed_args = mpm_new_argv->nelts;
     apr_getopt_init(&opt, process->pool, process->argc, process->argv);
     opt->errfn = NULL;
     optbuf[0] = '-';
